# HPROC2 -- Dispersion correct and linearize spectrum, placing input in Nodisp/
#           Correct for geometrially-caused pixel area variations
#           Correct for wavelength-dependent throughput variations
#           Divide extracted multispec spectra by sky absorption spectra
# By Doug Mink, SAO Telescope Date Center
# June 30, 2005

procedure hproc2 (images)

string	images = ""	{prompt="Input multispec image(s) to correct"}
string	calfile=""	{prompt="Calibration file for dispersion correction"}
real wl1=3700.0		{prompt="Starting wavelength for output spectra (A)"}
real wl2=9150.0		{prompt="Ending wavelength for output spectra (A)"}
real dwl=1.21	{prompt="Delta wavelength per pixel for output spectra (A)"}
int npix = INDEF	{prompt="Number of pixels per output spectra"}
string	skycorr=""	{prompt="Transmission from sky flat via maketran"}
string	skyfile="hectospec$lib/skyabs.fits"	{prompt="Sky absorption spectrum normalized to continuum=1"}
string	apertures=""	{prompt="List of input apertures or columns/lines"}
string	bands=""	{prompt="List of input bands or lines/bands"}
bool deltemp = no	{prompt="Delete temporary files (y or n)"}
bool saveold = yes	{prompt="Save original files in Nodisp/ and Nonlin/ (y or n)"}
bool verbose = yes	{prompt="Log progress (y or n)"}
bool debug = no		{prompt="Too much info from sumspec (y or n)"}

begin
	bool vb, db, save
	file imin, imlin, imdisp, imroot, imcal, newfile, templist
	file imout, imcorr, geores, geotemp, imnorm, imsky
	string ims, tchar, history, keypath, ap
	int lroot, lref
	string	imtype, mstype, keyroot, apxkey, apykey, rratio, aps
	string imageap, order, function
	int index, iap, nap, npix, lroot, xcol, ycol
	real x, y, rad, ratio

	ims = images
	imcal = calfile
	vb = verbose
	db = debug
	save = saveold
	imcorr = skycorr
	ap = apertures
	imsky = skyfile

	if (ims == "") {
	    print ("HPROC2:  Dispersion-correct and linearize spectra")
	    print ("HPROC2:  Divide extracted multispec spectra by sky absorption spectrum")
	    print ("Usage: hproc2 <input images>")
	    return
	    }

# Remove extension from calibration file, if one is present
	smatch (imcal, ".fits")
	if (smatch.found) {
	    lref = smatch.start - 1
	    imcal = substr (imcal, 1, lref)
	    }

#  Make temporary list of input files
	tchar = substr (ims,1,1)
        if (tchar == '@') {
	    templist = substr (ims,2,100)
	    deltemp = no
	    }
	else {
	    templist = mktemp ("linlist")
	    files (ims, >templist)
	    }
	list = templist
	while (fscan (list,imin) != EOF) {

# Make output file name and check to make sure input is multispec
	    smatch (imin, ".ms")
	    if (smatch.found) {
		lroot = smatch.start - 1
		imroot = substr (imin, 1, lroot)
		}
	    else {
		print ("HPROC2: ", imin," is not a multispec file")
		return
		}

# Make sure filename ends in .fits
	    smatch (imin, ".fits")
	    if (!smatch.found) {
		imin = imin // ".fits"
		}

# Do not linearize image twice
	    imgetp.default = "0"
	    imgetp (imin, "HLIN")
	    if (imgetp.value != "0") {
		if (vb) {
		    print ("HPROC2: ", imin, " has already been linearized")
		    print ("HLIN = ", imgetp.value)
		    }
		next
		}

# Dispersion-correct file
	    if (vb) {
		print ("HPROC2: Set REFSPEC1 in ",imin," to ",imcal)
		}
	    hedit (imin, "REFSPEC1", imcal, verify=no,show=vb,add=yes,update=yes)
#	    refspec (imin, references=imcal)
	    imdisp = imroot // ".disp.ms.fits"
	    dispcor (imin, imdisp, linearize=no)

# Save original file, if requested
	    if (save) {
		if (!access ("Nodisp/")) {
		    mkdir ("Nodisp")
		    }
		newfile = "Nodisp/" // imin
		if (access (newfile)) {
		    imdelete (newfile)
		    }
		move (imin, "Nodisp")
		if (vb) {
		    print ("HPROC2: Original ", imin, " => Nodisp/")
		    }
		}
	    else {
		delete (imin)
		}
	    rename (imdisp, imin)

# Make output file name
	    imlin = imroot // ".lin.ms.fits"

# Rebin spectra identically
	    rvsao (motd=no)
	    sumspec.spectra = imin
	    sumspec.compfile = imlin
	    sumspec.specnum = "1-300"
	    sumspec.specband = 0
	    sumspec.specdir = ""
	    sumspec.compname = ""
	    sumspec.compdir = ""
	    sumspec.nspec = 300
	    sumspec.save_names = no
	    sumspec.normin = 0.
	    sumspec.fixbad = "no"
	    sumspec.cont_remove = "no"
	    sumspec.reject = no
	    sumspec.abs_reject = 100.
	    sumspec.em_reject = 2.
	    sumspec.contout = "no"
	    sumspec.cont_plot = no
	    sumspec.cont_add = 0.
	    sumspec.spec_smooth = 0
	    sumspec.st_lambda = wl1
	    sumspec.end_lambda = wl2
	    sumspec.pix_lambda = dwl
	    sumspec.npts = npix
	    sumspec.complog = no
	    sumspec.interp_mode = "spline3"
	    sumspec.normout = 0.
	    sumspec.spec_plot = no
	    sumspec.spec_int = no
	    sumspec.comp_plot = vb
	    sumspec.comp_int = no
	    sumspec.velcomp = INDEF
	    sumspec.zcomp = INDEF
	    sumspec.svel_corr = "none"
	    sumspec.nsmooth = 0
	    sumspec.device = "stdgraph"
	    sumspec.plotter = "stdplot"
	    sumspec.logfiles = "STDOUT,displin.log"
	    sumspec.nsum = 1
	    sumspec.debug = db
	    sumspec.cursor = ""
	    sumspec (imin, imlin)

# Move original file to Nonlin directory, if requested
	    if (save) {
		if (!access ("Nonlin/")) {
		    mkdir ("Nonlin")
		    }
		newfile = "Nonlin/" // imin
		if (access (newfile)) {
		    imdelete (newfile)
		    }
		move (imin, "Nonlin")
		if (vb) {
		    print ("HLIN: Nonlinear ", imin, " => Nonlin/")
		    }
		}
	    else {
		delete (imin)
		}

# Rename result file to original file name
	    rename (imlin, imin)

# Add history line including time
	    gettime ()
	    history = gettime.date // " " // gettime.time // " image linearized"
            hedit (imin, "HLIN", history, add+, verify-, show-)


# Correct pixel fluxes for variation in pixel size due to projection effects
	    imout = substr (imin, 1, lroot) // ".corr.ms.fits"

	    if (vb) {
		print ("HPROC2: ", imin," -> ",imout)
		}

# Get number of apertures and pixels from header of input multispec image
	    imgetp.verbose = db
	    imgetp.default = "0"
	    imgetp (imin, "i_naxis1")
	    npix = int (imgetp.value)
	    imgetp (imin, "i_naxis2")
	    nap = int (imgetp.value)
	    if (vb) {
		print ("HPROC2: ", nap," apertures of ",npix," pixels each")
		}

# Check to see whether geometrical flux correction has been done
	    imgetp.verbose = db
	    imgetp (imin, "HGEOM")
	    if (imgetp.value == "0") {
		imgetp (imin, "FIBRGEOM")
		}
	    if (imgetp.value != "0") {
		if (vb) {
		    print ("HPROC2: ", imin, " has been geometriclly corrected")
		    print ("HGEOM = ", imgetp.value)
		    }
		next
		}

# Set up temporary geometric correction file
	    geotemp = mktemp ("geocorr") // ".ms.fits"
	    if (vb) {
		print ("HPROC2: radius normalization file is ", geotemp)
		}
	    chpixtype (imin, geotemp, "real")

# Compute geometric correction for each fiber based on its position
	    for (iap = 1; iap <= nap; iap = iap + 1) {
		if (iap == 1) {
		    apxkey = "APID" // iap // ",6"
		    imgetp.default = ""
		    imgetp (imin, apxkey)
		    if (imgetp.value == "") {
			xcol = 3
			ycol = 4
			}
		    else {
			xcol = 6
			ycol = 7
			}
		    }
		apxkey = "APID" // iap // "," //xcol
		imgetp.default = 9999.0
		imgetp (imin, apxkey)
		x = real (imgetp.value)
		apykey = "APID" // iap // "," //ycol
		imgetp (imin, apykey)
		y = real (imgetp.value)
		rad = sqrt (x * x + y * y)
		ratio = 1.11748 + (0.0001866 * rad) - (0.00000188622 * rad * rad);
		rratio = ratio
		imageap = geotemp // "[*," // iap // "]"
		if (db) {
		    print ("HGEOM: ", imin,"[",iap,"]: ", x, y, rad, " => ", ratio)
		    }
		imreplace (imageap, ratio)
		}

# Correct spectra for optical projection differences
	    geores = mktemp ("geores") // ".ms.fits"
	    imarith (imin, "/", geotemp, geores, verbose=yes)
	    if (deltemp) {
		delete (geotemp)
		}

# Add history line including time about geometry correction
	    gettime ()
	    history = gettime.date // " " // gettime.time // " geometric correction made"
	    hedit (geores, "HGEOM", history, add+, verify-)

# Move original file to Nogeo directory
	    if (saveold) {
		if (!access ("Nogeo/")) {
		    mkdir ("Nogeo")
		    }
		newfile = "Nogeo/" // imin
		if (access (newfile)) {
		    imdelete (newfile)
		    }
		move (imin, "Nogeo")
		}
	    else {
		delete (imin)
		}
	    imrename (geores, imin)

	    imout = substr (imin, 1, lroot) // ".corr.ms.fits"
	    if (vb) {
		print ("HPROC2: ", imin," -> ",imout)
		}

# If geometrical flux correction has not been done, do it
	    imgetp.verbose = db
	    imgetp (imin, "HGEOM")
	    if (imgetp.value == "0") {
		imgetp (imin, "FIBRGEOM")
		}
	    if (imgetp.value == "0") {
		heom (imin)
		}

# Check to see whether fiber color throughput correction has been done
	    imgetp.default = "0"
	    imgetp (imin, "HTRAN")
	    if (imgetp.value == "0") {
		imgetp (imin, "FIBRTRAN")
		}
	    if (imgetp.value != "0") {
		if (vb) {
		    print ("HPROC2: ", imin, " has been throughput corrected")
		    print ("HTRAN = ", imgetp.value)
		    }
		next
		}
	    if (imcorr == "") {
		if (vb) {
		    print ("HPROC2: ", imin, " has no throughput correction file")
		    }
		next
		}

# Correct all spectra for fiber color throughput differences
	    imarith (imin, "/", imcorr, imout, verbose=yes)

# Move original file to Notran directory
	    if (saveold) {
		if (!access ("Notran/")) {
		    mkdir ("Notran")
		    }
		newfile = "Notran/" // imin
		if (access (newfile)) {
		    imdelete (newfile)
		    }
		move (imin, "Notran")
		}
	    else {
		delete (imin)
		}

# Rename result file to original file name
	    imrename (imout, imin)

# Add history lines including time about this throughput correction file
	    gettime ()
	    imgetp (imcorr, "TRANORD")
	    order = imgetp.value
	    imgetp (imcorr, "TRANFUNC")
	    function = imgetp.value
	    gettime ()
	    history = gettime.date // " " // gettime.time // " fiber throughput corrected: " // order // " order " // function
	    hedit (imin, "HTRAN", history, add+, verify-)
	    hedit (imin, "FIBRFILE", imcorr, add+, verify-)

# Do not remove sky absorption twice
	    imgetp.default = "0"
	    imgetp (imin, "HABSKY")
	    if (imgetp.value != "0") {
		if (vb) {
		    print ("HABSKY: ", imin, " has already been processed")
		    print ("HABSKY = ", imgetp.value)
		    }
		next
		}

# Copy original file to Noskyab directory
	    if (saveold) {
		if (!access ("Noskyab/")) {
		    mkdir ("Noskyab")
		    }
		newfile = "Noskyab/" // imin
		if (access (newfile)) {
		    imdelete (newfile)
		    }
		imcopy (imin, "Noskyab")
		if (vb) {
		    print ("HABSKY: Original ", imin, " => Noskyab/")
		    }
		}

# Divide out normalization spectra
	    imout = mktemp ("abskyout") // ".ms.fits"
	    sarith.w1 = INDEF
	    sarith.w2 = INDEF
	    sarith.apertures = ap
	    sarith.bands = bands
	    sarith.beams = ""
	    sarith.apmodulus = 0
	    sarith.reverse = no
	    sarith.ignoreaps = yes
	    if (ap == "") {
		sarith.format = "onedspec"
		}
	    else {
		sarith.format = "multispec"
		}
	    sarith.renumber = no
	    sarith.offset = 0
	    sarith.clobber = yes
	    sarith.merge = no
	    sarith.rebin = yes
	    sarith.errval = 0.
	    sarith.verbose = vb
	    sarith (imin, "/", imsky, imout)

# Fix APNUM keywords
	    keypath = "@" // osfn("hectospec$lib/") // "apnumchg"
	    print ("keyhead ",imout," ",keypath)
	    keyhead (imout,keypath)
	    keypath = "@" // osfn("hectospec$lib/") // "apidkeys"
	    print ("cphead ",imin," ",imout," ",keypath)
	    chead (imin,imout,keypath)

# Add history line including time
	    gettime ()
	    history = gettime.date // " " // gettime.time // " " // imsky // " divided out"
	    hedit (imout, "HABSKY", history, add+, verify-, show-)

# Rename output file to same as input file (which is saved if saveold=yes)
	    imdelete (imin)
	    imcopy (imout, imin)
	    }

	if (deltemp) {
	    delete (templist)
	    }
end
# Jun 10 2004	New task
# Sep  2 2004	Run on multiple files via list argument or file
# Sep  2 2004	Add keyword HLIN and do not linearize image when it already exists
# Sep 23 2004	Change "transmission" to "throughput"
# Oct 13 2004	Drop extra closing brace
# Oct 14 2004	Write sarith output to new file and rename it
# Oct 14 2004	Replace sarith-written APNUMnnn with APIDnnn from input file
# Oct 18 2004	If already processed, print HLIN value
# Oct 18 2004	Change FIBRTRAN keyword to HTRAN
# Oct 18 2004	Print value of HTRAN/FIBRTRAN if already processed
# Oct 18 2004	Split off geometric correction to HGEOM task
# Oct 18 2004	Print HABSKY value if already processed

# Jun 30 2005	Change external cphead call to wcstools.chead
