#!/data/oir/bin/ksh
#
# Generate starbase tables of:
#
#   1) All the keywords in all FITS files     --> ${dirname}_keyval.db
#   2) All Object files                       --> ${dirname}_objlist.db
#   3) All Skyfiles & Certification values    --> ${dirname}_skylist.db
#   3) Objects and distances to each Sky file --> ${dirname}_skydist.dd
#
# Number (3) uses keywords FIBER_X and FIBER_Y to compute the distance
# between an object and each of the sky files in turn.
#
# Also, assumes the UNIX IRAF environment is set up (for imhbase, and all
# the other IRAF utilities).
#
#  
#
# Produces a <dir>_skydist.db starbase file that looks like (example):
# HDRFILE   Skyfile   FIBER_X FIBER_Y SkyX    SkyY    SkyDist SkyR
# --------- --------- ------- ------- ------- ------- ------- ----
# obj.fits  sky4.fits  10.000  10.000  70.000  11.000    60.0  1.8
# obj.fits  sky3.fits  10.000  10.000 -50.000  40.000    67.1  2.2
# obj2.fits sky4.fits  20.000  60.000  70.000  11.000    70.0  1.8
# obj2.fits sky3.fits  20.000  60.000 -50.000  40.000    72.8  2.2
#
#

# ========================================================================
#       Setup the Korn Shell envirnment for variables and parameter files
# ========================================================================

export home=${irafhome:-$HOME/unixiraf}
export DEFAULT_DIR=${iraf}../iraf.pkg/hectospec/
export PATH=/data/oir/unixiraf12.1/bin:$PATH

export PATH=/data/oir/starbase2.7.8:$PATH

. /data/oir/unixiraf12.1/scripts/ui_setup.ksh 
. $ENVIRONMENT

# ========================================================================
#       Get runtime parameters from script parameter file
# ========================================================================
#

typeset -F FiberRadiusLimit=306.0 # limit of non-parked fibers

verbose=0
destdir="skysub/"
files=\*.fits
mpjobs=12
nskys=0
fiberdist=0.
logfile=skyproc.results
skycert=4.
skystdlim=0.2
skysame=1
emline=2
uparmdir=${uparm:-$UPARM}

while getopts "c:d:e:f:i:l:m:n:r:s:u:v:hx" fval ; do
   case $fval in
    c) if   [[ $OPTARG == yes ]]; then skysame=1
       elif [[ $OPTARG == no  ]]; then skysame=0
       elif [[ $OPTARG -ne 0  ]]; then skysame=1
       elif [[ $OPTARG -eq 0  ]]; then skysame=0
       fi;;
    d) destdir=$OPTARG   ;;
    e) emline=$OPTARG    ;;
    i) files=$OPTARG     ;;
    m) mpjobs=$OPTARG    ;;
    n) nskys=$OPTARG     ;;
    f) fiberdist=$OPTARG ;;
    l) logfile=$OPTARG   ;;
    r) skycert=$OPTARG   ;;
    s) skystdlim=$OPTARG ;;
    u) uparmdir=$OPTARG  ;;
    v) verbose=$OPTARG   ;;
   h|x) print  "Usage: $(basename $0) [options]"
        printf "  -i %s\tinput file list\n" "$files"
        printf "  -d %s\tdestination directory for output files\n" $destdir
        printf "  -n %d\t\tMax number of skys to coadd\n" $nskys
        printf "  -f %.2f\tMax fiber distance to sky\n" $fiberdist
        printf "  -r %.2f\tMax R value for certified sky\n" $skycert
        printf "  -s %.2f\tMax std of difference from avg sky\n" $skystdlim
        printf "  -c %d\t\tUse skys only from same chip: %s\n" $skysame \
                "$(if [[ $skysame -gt 0 ]]; then echo yes; else echo no;fi)"
        printf "  -e %d\t\tEmission line to scale by: 1=O5577, 2=OH8399\n" \
                         $emline
        printf "  -l %s\tName of output summary file\n" $logfile
        printf "  -m %d\t\tMax number of multiprocessor jobs in parallel\n" \
                         $mpjobs
        fval=${uparmdir:-$UPARM}
        printf "  -u %s\tName of uparm directory\n" "$fval"
        printf "  -v %d\t\tverbose - output summary (on stderr):\n" \
                         $verbose 
        printf "\t\t  0=quiet, 1=notes, 2=very verbose, 3=watch out\n"
        printf "  -(h|x)\t\tprint this usage and exit\n"
        printf  "\n"
        exit 1;;
#    \?) print -u2 "Unknown option $OPTARG"; exit 1;;
     :) print -u2 "-$OPTARG requires an argument"; exit 1;;
   esac
done

if (( mpjobs > 4096 )) then
    mpjobs = 4096
elif (( mpjobs < 1 )) then
    mpjobs = 1
    fi

# Make sure output directory is there
destdir=`pwd`/$destdir
if [[ -d $destdir ]] then
    x=0
else
    mkdir -p $destdir
fi

trap exit ERR
export archive=$irafpkg/saotdc/pkg/archive/
export imextn="oif:imh fxf:fits,fit,fts plf:pl qpf:qp stf:hhh,??h"

skymethod="SKYSUB"

REV=2.1m

tmp="/tmp/$(basename $0).$$"/
mkdir -p $tmp

tmpfileA=${tmp}$$A
tmpfileB=${tmp}$$B
tmpskylist=${tmp}$$.skyhdr

procpid=$$

exitval=0

trap 'print -u2 -f "ERROR $? in $0 at line $LINENO\n"; exitval=1; exit' ERR

trap "/bin/rm -rf $tmp" EXIT INT TERM HUP 

# Initialize parameter files from hectospec package if not in uparm directory
# and copy to destination directory for use by subroutines
#
# E.g. home$uparm/hecskemso.par  -> $destdir/emsao.par
#  OR
#      $iraf.pkg/hectospec/skemsao.par -> 
#           home$uparm/hecskemso.par and also -> $destdir/emsao.par
#
getirafparm emsao        skyproc hectospec $uparmdir $destdir # hecskemso.par
getirafparm imarith      skyproc hectospec $uparmdir $destdir # hecskimah.par
getirafparm imcombine    skyproc hectospec $uparmdir $destdir # hecskimce.par
getirafparm imdelete     skyproc hectospec $uparmdir $destdir # hecskimde.par
getirafparm imexpr       skyproc hectospec $uparmdir $destdir # hecskimer.par
getirafparm imrename     skyproc hectospec $uparmdir $destdir # hecskimre.par
getirafparm imstack      skyproc hectospec $uparmdir $destdir # hecskimsk.par
getirafparm imstatistics skyproc hectospec $uparmdir $destdir # hecskimss.par
getirafparm xcsao        skyproc hectospec $uparmdir $destdir # hecskxcso.par

# now make $destdir == the unixiraf UPARM dir

uparmdir=$destdir
export UPARM=$uparmdir
if (( verbose > 0 )) then
    echo Parameter files are in $UPARM
fi

# String prefacing output .db files - default is the directory name.
prefix=${prefix:-$(basename $(pwd))_}
if [[ $prefix == "NONE" ]]; then prefix=""; fi

if [[ -z "$logfile" ]] then  logfile="/dev/null"; fi

if [[ $logfile != "-" && $logfile != "STDOUT" ]] then
  exec 1>> $logfile
fi

if (( verbose > 0 )) then
    printf "\n%s rev %s start at $(date)\n\n" $(basename $0) $REV
    printf "  FiberRadiusLimit = %.1f\n" $FiberRadiusLimit
    printf "  Nskys = %d\n  FiberDist = %.1f\n  SkyCert <= %.1f\n" \
              $nskys $fiberdist $skycert
    printf "  SkyStdLim = %.2f\n" $skystdlim
    printf "  Skys on same chip = %s\n" \
      $(if [[ $skysame -ge 1 ]]; then echo yes; else echo no; fi)
    printf "  DB Files prefix = %s\n" $prefix
    printf "  DestDir = %s\n\n" $destdir
fi


# Generate master table of keywords; works for FITS or IRAF files
files=$(echo "$files" |tr "," " ")
eval gethead -tha IMAGETYP OBJECT APERTURE   \
           FIBER FIBER_X FIBER_Y "$files"   |
  trim   FILENAME=HDRFILE -a  CHIPNUM FIBER_RADUIS                       |
  compute 'if(OBJECT ~ /_-1$/) {OBJECT = "UNUSED"; IMAGETYP = "UNUSED" }
           if(OBJECT ~ /_-2$/) {OBJECT = "BROKEN"; IMAGETYP = "BROKEN" }
           if(OBJECT ~ /_0$/ ) {OBJECT = "SKY"}
           FIBER_RADIUS = sqrt(square(FIBER_X) + square(FIBER_Y))
           if(APERTURE <= 150) { CHIPNUM = 1 * '$skysame'
             } else { CHIPNUM = 2 * '$skysame' }
  ' |
  sorttable HDRFILE >${prefix}keyval.db

# Pull all [object and not sky] files into a master list
row <${prefix}keyval.db \
   'IMAGETYP == "OBJECT" && toupper(OBJECT) !~ /(UNUSED|PARKED|BROKEN)/ && 
        FIBER_RADUIS <= '$FiberRadiusLimit   |
column HDRFILE OBJECT CHIPNUM FIBER FIBER_X FIBER_Y    \
>${prefix}objlist.db 

# create a sky file db
row 'FIBER_RADUIS <= '$FiberRadiusLimit' && 
     IMAGETYP == "OBJECT" && toupper(OBJECT) ~ "SKY"' <${prefix}keyval.db |
column HDRFILE CHIPNUM | sorttable CHIPNUM >${tmp}sky.db
nsky=$(wc -l <${tmp}sky.db)
if [[ nsky -le 2 ]]
   then
   echo "No sky spectra in this directory"
   exit -1
   fi
cat ${tmp}sky.db
# create the sky file list (not a .db file).
chipval="."   
if [ skysame -gt 0 ]; then chipval="1 2"; fi
{
 for chip in $chipval; do

  avout=${tmp}skyav.${chip}.fits
  row 'CHIPNUM ~ /'$chip'/' <${tmp}sky.db | headoff HDRFILE  >$tmpskylist
  nskys=$(wc -l <$tmpskylist)
  /bin/rm -f $avout

  # compute master sky average, then ratio each against the average to pick
  # up abberant skys to remove from the master sky list
#  tmprat=${tmp}skyratio
  sed <$tmpskylist -e 's/\.sky_.*\./.ratio./' -e "s,^,$tmp," >$tmpfileA

  imcombine @$tmpskylist $avout rejmask="" combine="average"          \
        reject="avsigclip"   outtype="real"  scale="none"   zero="none"   \
        weight="!EXPTIME"  mclip="yes"  lsigma=3. hsigma=3. sigscale=0.1  \
        logfile=""
  imarith @$tmpskylist / $avout @$tmpfileA
  imstatistics @$tmpfileA fields=mean,stddev format=no  |
  while read stddev; do 
    printf "%d\t%s\t%s\t%s\n" $nskys $mean $stddev $avout
  done
  /bin/rm -f $(cat $tmpfileA)

 done
} | 
fldtotable nskys skyavmean skyavstd avskyfile > ${tmp}skyav.db 

pastetable ${tmp}sky.db ${tmp}skyav.db     |
sorttable HDRFILE >${tmp}skyavstd.db

# generate the sky files each minus the average sky
headoff <${tmp}skyavstd.db HDRFILE >$tmpskylist
headoff <${tmp}skyavstd.db avskyfile >$tmpfileA
/usr/bin/sed -e 's/\.sky_.*\./.diff./' -e "s,^,$tmp," <$tmpskylist >$tmpfileB
imarith @$tmpskylist - @$tmpfileA @$tmpfileB

# cross-correlate each
(
 mkdir -p ${tmp}xc1  2>/dev/null
  cp $destdir/xcsao.par ${tmp}xc1
  UPARM=${tmp}xc1; export UPARM
  xcsao @$tmpfileB report_mode=-8 displot=no templates=femtemp97.imh |
  column xcr=SkyRem >${tmp}skyxem.db
) &

(
 mkdir -p ${tmp}xc2  2>/dev/null
  cp $destdir/xcsao.par ${tmp}xc2
  UPARM=${tmp}xc2; export UPARM
  xcsao @$tmpfileB  report_mode=-8 displot=no mode=q templates=fabtemp97.imh|
  column xcr=SkyRab >${tmp}skyxab.db
) &
wait
pastetable ${tmp}skyavstd.db ${tmp}skyxem.db ${tmp}skyxab.db |
column -a SkyR -x avskyfile                                  |
compute 'SkyR = SkyRem; if(SkyRab > SkyRem) { SkyR = SkyRab }' \
 >${prefix}skylist.db

nsky=$(wc -l <${prefix}skylist.db)
if [[ nsky -le 2 ]]
   then
   echo "No usable sky spectra in this directory"
   exit -1
   fi

# Pull join (cprodtable) all certified skyfiles to object list, compute 
# distance from each object to each sky, leave sorted by object and
# then sky distances.

cprodtable  ${prefix}objlist.db                                      \
   <(jointable -j HDRFILE ${prefix}keyval.db ${prefix}skylist.db  |
      column HDRFILE CHIPNUM_1=CHIPNUM FIBER FIBER_X FIBER_Y    \
               nskys=Nskys SkyR skyavstd=SkyAvStd               |
      row '(SkyR ~ /^ *$/ || SkyR <= '$skycert') &&             
             SkyAvStd < '$skystdlim 
    ) |
    row 'CHIPNUM_1 == CHIPNUM_2'                                  |
    column HDRFILE_1=HDRFILE HDRFILE_2=Skyfile                       \
          FIBER_1=FIBER FIBER_X_1=FIBER_X FIBER_Y_1=FIBER_Y          \
          FIBER_2=SkyFIBER FIBER_X_2=SkyX FIBER_Y_2=SkyY             \
          Nskys  SkyDist SkyR SkyAvStd                                |
    compute 'SkyDist=sprintf("%.1f",          \
                   sqrt(square(FIBER_X-SkyX) + square(FIBER_Y-SkyY)))' |
    sorttable HDRFILE SkyDist:n  \
 >${prefix}skydist.db

# just to see the final table
[[ $verbose -gt 2 ]] &&
  justify <${prefix}skydist.db                                 \
      HDRFILE,7 Skyfile,7 FIBER,4 FIBER_X,8 FIBER_Y,8          \
      SkyFIBER,6 SkyX,8 SkyY,8 SkyDist,7 SkyR,4 

[[ $verbose -gt 0 ]] && {
   printf "\nElapsed time after object & sky tables creation: $SECONDS seconds\n"
   times
}

# run emsao on list of object files to get strength of the selected
# sky line (O5577 or OH8399) line, and add that column to table.
pset emsao report_mode=-4 displot=no esmooth=2
if [[ $emline -eq 1 ]]; then
  pset emsao emsearch=searchO5577.dat emlines=O5577.dat linedir='/data/IRAF2.12/iraf.pkg/rvsao/lib/'
  emlabel="O5577"
else
  pset emsao emsearch=searchOH8399.dat emlines=OH8399.dat linedir='/data/IRAF2.12/iraf.pkg/rvsao/lib/'
  emlabel="OH8399"
fi
headoff HDRFILE <${prefix}objlist.db >$tmpfileA
#echo Files from ${prefix}objlist.db in $tmpfileA
seqno=0
nsplit=$(( ($(wc -l <$tmpfileA) / mpjobs) + 1))
split -l $nsplit $tmpfileA ${tmp}objpart.

for objpart in ${tmp}objpart.*; do
  seqno=$((seqno + 1))
  (
    xit=0
    tdir=${tmp}objdir.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $UPARM/emsao.par $tdir
    export UPARM=$tdir
#    cp $objpart /tmp/objpart

#    echo emsao @$objpart report_mode=-4 displot=no esmooth=2 mode=q

    emsao @$objpart report_mode=-4 displot=no esmooth=2 mode=q  | \
    column specid=HDRFILE lht01=objEmiss  >$tmpfileB.$seqno >${objpart}.db

    xit=$?
    [[ $xit != 0 ]] && {
      print "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
    }
    exit $xit
  ) &  # END FORKED SUBSHELL
done  
# wait for the subprocesses to finish
wait

# Pull the sky tables all together
if [[ $(echo ${tmp}objpart*.db | wc -w) -gt 1 ]]; then
  uniontable ${tmp}objpart*.db | sorttable HDRFILE >$tmpfileB
else
  sorttable HDRFILE <${tmp}objpart.aa.db >$tmpfileB
fi

jointable -j HDRFILE ${prefix}objlist.db $tmpfileB >$tmpfileA

if [[ $(wctable -l <$tmpfileA) -lt 1 ]]; then 
   mv $tmpfileB emsao.db
   printf "No rows in jointable?\n"
   exit 1
fi
mv $tmpfileA ${prefix}objlist.db
/bin/rm -rf ${tmp}objpart.* ${tmp}objdir.* $tmpfileB.*

[[ $verbose -gt 0 ]] && {
   printf "\nElapsed time after object emission line fit: $SECONDS seconds\n"
   times
}

#
# Create entries in table for sky summation list for each object
#
seqno=0
waitjob=0
/bin/rm -f ${tmp}nskys.*

for obj in $(headoff <${prefix}objlist.db HDRFILE); do
   seqno=$((seqno + 1))
 (
   tmpA=${tmpfileA}.${seqno}

# Select all skyfiles that are certified and within FIBERDIST of object.
# Not that the test for samechip status (if enabled) has already been done.
   trim <${prefix}skydist.db    |
   table -h '
     HDRFILE == "'$obj'"                             && \
     (SkyR <= '$skycert' || SkyR ~ /^ *$/)           && \
     ('$fiberdist' == 0  || SkyDist <= '$fiberdist') {print Skyfile}
   ' >$tmpA

# Limit to the max allowed number of skys. This chooses the closet ones 
# because the input table was sorted secondarily by SkyDist.
  if [[ $nskys -gt 0 ]]; then
    head -${nskys} $tmpA      |
    sort >${tmpskylist}.$seqno
  else
    sort <$tmpA >${tmpskylist}.$seqno
  fi

# ERROR if no sky files left after selection
  if [[ ! -s ${tmpskylist}.$seqno ]]; then    
    print -u2 -f "\nWARNING:" 
    print -u2 -f " File %s has no sky files within the %.3f mm limit!\n"   \
       $obj $fiberdist
    exitval=1
    continue
  fi
  nskyfiles=$(wc -l <${tmpskylist}.$seqno)
  skylist=$(tr '\012' '+' <${tmpskylist}.$seqno)
  skylist=${skylist%+}
  aplist=$(gethead -a @${tmpskylist}.$seqno aperture | 
            cut -f 2 -d\  | 
            tr '\012' '+')
  printf "%s\t%s\t%d\t%s\n" $obj ${skylist%_} $nskyfiles ${aplist%+} \
       >>${tmp}nskys.$seqno
 ) &
  if [[ $((seqno - waitjob)) -ge $mpjobs ]]; then
      waitjob=$((waitjob + 1))
      wait %${waitjob}
  fi
 done 

 wait 

 cat ${tmp}nskys.*                                |
 fldtotable HDRFILE Skyfilelist Nskys Apertures   | 
 sorttable HDRFILE >$tmpfileB

jointable -j HDRFILE ${prefix}objlist.db $tmpfileB  >$tmpfileA

mv $tmpfileA ${prefix}objlist.db

/bin/rm -f ${tmp}nskys.*   ${tmpskylist}.*

[[ $verbose -gt 1 ]] && {
  printf "\nSet of skys for object files:\n"
  justify <${prefix}objlist.db HDRFILE Nskys Apertures
}

#
# Create sorted, unique set of skyfile lists with emission line peak value
#
pset emsao report_mode=-4 displot=no esmooth=2
pset imcombine project=no logfile=""

skynum=0

headoff Skyfilelist <${prefix}objlist.db | sort -u >${tmp}skyfilelist

seqno=0
nsplit=$(( ($(wc -l <${tmp}skyfilelist) / mpjobs) + 1))
split -l $nsplit ${tmp}skyfilelist ${tmp}skyflist.

for skylist in ${tmp}skyflist.*; do
  seqno=$((seqno + 1))
  skynum=0
  /bin/rm -rf ${skylist}.db

# fork off for multiproc
  (
    xit=0
    tdir=${skylist}.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $UPARM/emsao.par  $UPARM/imcombine.par $tdir
    export UPARM=$tdir
#
# Get sky fiber list from sky file names
#
    cat $skylist   |
    while read skyfiblist; do
      skynum=$((skynum + 1))
      nskyfibers=$(echo $skyfiblist | tr "+" "\012" | wc -l)
      skyfibers="$(    \
       (for file in $(echo $skyfiblist | tr '+' ' '); do
          table -h <${prefix}keyval.db 'HDRFILE == "'$file'" {print FIBER}'
        done) | tr "\012" "_")"

      extn=${skyfiblist##*\.}
      extn=${extn%%+}
      skysumfile=${tmp}${skyfibers}.${procpid}.${extn}
      shortname=${tmp}skyset_${seqno}_${skynum}.$extn
#  shortname=skyset_${seqno}_${skynum}.$extn
      /bin/rm -f $skysumfile $shortname
      echo $skyfiblist | tr '+' '\n' >${tdir}skyfiles

      imcombine @${tdir}skyfiles $shortname mode=q
      skyEmiss=$(emsao $shortname mode=q |  headoff lht01)

      xit=$?
      [[ $xit != 0 ]] && {
        print "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
      }
      printf "%s\t%s\t%s\t%s\t%s\t%s\n" $skyfiblist $skysumfile $shortname \
         "$nskyfibers" $skyfibers $skyEmiss >>${skylist}.db
    done
  ) &
done

wait

cat ${tmp}skyflist*.db |
fldtotable Skyfilelist Skysumfile Skyabbrev Nskyfibers Skyfibers skyEmiss |
sorttable Skyfilelist >$tmpfileA

/bin/rm -rf ${tmp}skyfilelist ${tmp}skyflist*

#cp ${prefix}objlist.db /tmp/objlist.db
#cp $tmpfileA /tmp/fileA
  
jointable -j Skyfilelist <(sorttable Skyfilelist <${prefix}objlist.db) \
                          $tmpfileA                                     |
 column -a ratio                                                        |
 compute 'ratio=sprintf("%.6f", objEmiss/skyEmiss)'  >$tmpfileB

sorttable HDRFILE <$tmpfileB >${prefix}objlist.db

[[ $verbose -gt 0 ]] && {
   printf "\nElapsed time after summed sky emission line fit: $SECONDS seconds\n"
   times
}

# Make sure output directory is there
[[ -z "$destdir" || -d $destdir ]] || mkdir -p $destdir

# Subtract off  scaled sky
#typeset -Z5 seqno=0
seqno=0

[[ $verbose -gt 1 ]] && printf "\n"

columns=\
"HDRFILE Skysumfile Skyabbrev Nskyfibers Skyfibers Apertures objEmiss skyEmiss ratio"

mpjobs=$(($mpjobs > 4096 ? 4096 : $mpjobs))
nxtjob=0
waitjob=0

headoff $columns <${prefix}objlist.db |&
while read -p $columns; do
  seqno=$(($seqno + 1))

# Fork off up to $mpjobs subshells into the background, waiting for
# one before doing the next
 (  
  temp=${tmp}$seqno/
  trap \
   'print -u2 "B:ERROR $? in $0 seq $seqno at line $LINENO"; exit 100' \
            ERR
  trap '/bin/rm -rf $temp' EXIT

  mkdir -p $temp 2>/dev/null

  [[ verbose -gt 2 ]] && 
   print -u2 -f "Forking skysub job %d for %s\n" $seqno $HDRFILE

  exec 1> ${tmp}logfile.$seqno 

  [ $verbose -gt 1 ] && { 
  printf "%s: Obj ${emlabel}= %.1f, Sky ${emlabel}= %.1f, ratio= %.3f\n"\
           $HDRFILE $objEmiss $skyEmiss $ratio
  }

  extn=$(expr $HDRFILE : '.*\(\..*\)$' \| $HDRFILE)
  outbase=$(basename $HDRFILE $extn)
  outfile=${temp}${outbase}.sub$extn
  destfile=${destdir}${outbase}.ms${extn}
  tmpdestfile=${temp}${outbase}.ms${extn}  

  #
  # for multiprocessing, need to make separate copy of the parameters file
  # so they don't conflict with each other
  
  cp ${UPARM}imexpr.par   ${UPARM}imstack.par \
     ${UPARM}imrename.par ${UPARM}imdelete.par  ${temp}

  UPARM=$temp
  export UPARM  

  imexpr 'a - b * c' $outfile $HDRFILE $Skyabbrev $ratio verbose=no mode=q \
        1>/dev/null
  exitval=$?

 # make 4505 by 1 files
  imstack $outfile ${temp}obj$extn
  imstack $HDRFILE ${temp}raw$extn
  imstack $Skyabbrev ${temp}sky$extn

 # now combine into one 4605 by 1 by 3 file
  imstack \
     "${temp}obj${extn},${temp}raw${extn},${temp}sky${extn}" $tmpdestfile 
  exitval=$?

  imdelete $outfile verify=no; xit=$?
  [[ $xit != 0 ]] && {
    print "$(basename $0): can't imdelete ${outfile}\?\n";  exit $xit
  }
  
  delhead BANDID1 BANDID2 BANDID3 BANDID4 $tmpdestfile
  now=`getdate 2fd now`
  datemsg=$(printf "Sky subtracted %s" $now)
  sethead BANDID1='spectrum - scaled sky'                        \
          BANDID2='raw spectrum'                                 \
          BANDID3='averaged sky'                                 \
          NSKYFIB=$Nskyfibers                                    \
          SKYFIB="$Skyfibers"                                    \
          SKYAPER="$Apertures"                                   \
          SKYCERT=$skycert                                       \
          NSKYMAX=$nskys                                         \
          SKYDIST=$fiberdist                                     \
          SKYMETHD="$skymethod"                                  \
          SKYRATIO=$ratio                                        \
          SKYSTDLM=$skystdlim                                    \
          SKYPROC=$datemsg                                       \
          $tmpdestfile 

 # Fix for when APNUM1 keyword not passed through
  apname=$(gethead APNUM1 $tmpdestfile)
  if [[ -z $apname ]]; then
    aperture=$(gethead APERTURE $tmpdestfile)
    [[ -n $aperture ]] && {
      apnum=$(printf "%s %s" $aperture $aperture)
      sethead $tmpdestfile APNUM1="$apnum"
    }
  fi

#  /bin/mv $tmpdestfile $destfile; xit=$?
#  [[ $extn = ".imh" ]] && {
#    /bin/mv $(basename $tmpdestfile .imh).pix $(basename $destfile $extn).pix
#    xit=$(($xit | $?))
#  }  
  imrename $tmpdestfile $destfile; xit=$?
  [[ $xit != 0 ]] && {
    print "$(basename $0): can't imrename ${outfile}\?\n"
    exit $xit
  }
  
  [[ $exitval -ne 0 ]] && print -u2 -f "ERROR: imexpr: status $exitval\n" 
  [[ $verbose -gt 1 ]] && { 
    printf "Output: %s\n\n" $outfile
  }

  exit $exitval
 ) &  # END FORKED SUBSHELL

  bkgnd[$nxtjob]=$!
  hdrfile[$nxtjob]=$HDRFILE
  nxtjob=$(($nxtjob+1)) 

# if mpjobs jobs are not running, continue forking before waiting
  [[ $(($nxtjob - $waitjob)) -lt $mpjobs ]] && continue
  
  wait ${bkgnd[$waitjob]}
  retval=$?

  retarr[$waitjob]=$retval
  [[ $retval -ne 0 ]] && {
     printf "\nERROR val %d while skyproc.sub of %s\n" \
             $retval ${hdrfile[$waitjob]}
     exitval=$retval
  }
  waitjob=$(($waitjob + 1))
  cat ${tmp}logfile.$waitjob
  /bin/rm -f ${tmp}logfile.$waitjob
done

# wait for last several to finish
while [[ $waitjob -lt $nxtjob ]] do
  wait ${bkgnd[$waitjob]}
  retval=$?
#  printf "Wait: %d %d ret = %d\n" $waitjob ${bkgnd[$waitjob]} $retval 
  retarr[$waitjob]=$retval
  [[ $retval -ne 0 ]] && {
     printf "\nERROR val %d while skyproc.sub of %s\n"   \
          $retval ${hdrfile[$waitjob]}
     exitval=$retval
  }
  waitjob=$(($waitjob + 1))
  cat ${tmp}logfile.$waitjob
done  

[[ $verbose -gt 0 ]] && {
#   [ "$logfile" != "-" ] && exec 1>> $logfile
   printf "\nElapsed time at exit: $SECONDS seconds\n"
   times
}

exit $exitval

# Aug 26 2004	DJM Process SKY spectra like OBJECT spectra
# Aug 26 2004	DJM Add SKYPROC keyword (will add test and drop out later)
# Oct 14 2004	DJM Add test for no sky files with error message and exit
# Nov 17 2004   WFW Add radius > 306 to reject parked fibers
